from typing import Optional, Callable, Any, Tuple, TypeVar

from ..model import Model
from ..util import use_nvtx_range


_ModelT = TypeVar("_ModelT", bound=Model)


def with_nvtx_range(
    layer: _ModelT,
    name: Optional[str] = None,
    *,
    forward_color: int = -1,
    backprop_color: int = -1,
) -> _ModelT:
    """Wraps any layer and marks the forward and backprop phases as
    NVTX ranges for CUDA profiling.

    By default, the name of the layer is used as the name of the range,
    followed by the name of the pass.
    """
    name = layer.name if name is None else name

    orig_forward = layer._func
    orig_init = layer.init

    def forward(model: Model, X: Any, is_train: bool) -> Tuple[Any, Callable]:
        with use_nvtx_range(f"{name} forward", forward_color):
            layer_Y, layer_callback = orig_forward(model, X, is_train=is_train)

        def backprop(dY: Any) -> Any:
            with use_nvtx_range(f"{name} backprop", backprop_color):
                return layer_callback(dY)

        return layer_Y, backprop

    def init(_model: Model, X: Any, Y: Any) -> Model:
        if orig_init is not None:
            return orig_init(layer, X, Y)
        else:
            return layer

    layer.replace_callbacks(forward, init=init)

    return layer



from typing import Tuple, Callable, Optional, TypeVar, Union, cast

from ..backends import NumpyOps
from ..model import Model
from ..config import registry
from ..types import Padded, Ragged, ArrayXd, Array3d, ListXd


NUMPY_OPS = NumpyOps()


ArrayTXd = TypeVar("ArrayTXd", bound=ArrayXd)
SeqT = TypeVar("SeqT", bound=Union[Padded, Ragged, ListXd, ArrayXd])


@registry.layers("with_array.v1")
def with_array(layer: Model[ArrayTXd, ArrayTXd], pad: int = 0) -> Model[SeqT, SeqT]:
    """Transform sequence data into a contiguous array on the way into and
    out of a model. Handles a variety of sequence types: lists, padded and ragged.
    If the input is an array, it is passed through unchanged.
    """
    model: Model[SeqT, SeqT] = Model(
        f"with_array({layer.name})",
        forward,
        init=init,
        layers=[layer],
        attrs={"pad": pad},
        dims={name: layer.maybe_get_dim(name) for name in layer.dim_names},
    )
    return model


def forward(
    model: Model[SeqT, SeqT], Xseq: SeqT, is_train: bool
) -> Tuple[SeqT, Callable]:
    if isinstance(Xseq, Ragged):
        return cast(Tuple[SeqT, Callable], _ragged_forward(model, Xseq, is_train))
    elif isinstance(Xseq, Padded):
        return cast(Tuple[SeqT, Callable], _padded_forward(model, Xseq, is_train))
    elif not isinstance(Xseq, (list, tuple)):
        return model.layers[0](Xseq, is_train)
    else:
        return cast(Tuple[SeqT, Callable], _list_forward(model, Xseq, is_train))


def init(
    model: Model[SeqT, SeqT], X: Optional[SeqT] = None, Y: Optional[SeqT] = None
) -> None:
    layer: Model[ArrayXd, ArrayXd] = model.layers[0]
    layer.initialize(
        X=_get_array(model, X) if X is not None else X,
        Y=_get_array(model, Y) if Y is not None else Y,
    )
    for dim_name in layer.dim_names:
        value = layer.maybe_get_dim(dim_name)
        if value is not None:
            model.set_dim(dim_name, value)


def _get_array(model, X: SeqT) -> ArrayXd:
    if isinstance(X, Ragged):
        return X.dataXd
    elif isinstance(X, Padded):
        return X.data
    elif not isinstance(X, (list, tuple)):
        return cast(ArrayXd, X)
    else:
        return model.ops.flatten(X)


def _list_forward(
    model: Model[SeqT, SeqT], Xs: ListXd, is_train: bool
) -> Tuple[ListXd, Callable]:
    layer: Model[ArrayXd, ArrayXd] = model.layers[0]
    pad = model.attrs["pad"]
    lengths = NUMPY_OPS.asarray1i([len(seq) for seq in Xs])
    Xf = layer.ops.flatten(Xs, pad=pad)
    Yf, get_dXf = layer(Xf, is_train)

    def backprop(dYs: ListXd) -> ListXd:
        dYf = layer.ops.flatten(dYs, pad=pad)
        dXf = get_dXf(dYf)
        return layer.ops.unflatten(dXf, lengths, pad=pad)

    return layer.ops.unflatten(Yf, lengths, pad=pad), backprop


def _ragged_forward(
    model: Model[SeqT, SeqT], Xr: Ragged, is_train: bool
) -> Tuple[Ragged, Callable]:
    layer: Model[ArrayXd, ArrayXd] = model.layers[0]
    Y, get_dX = layer(Xr.dataXd, is_train)

    def backprop(dYr: Ragged) -> Ragged:
        return Ragged(get_dX(dYr.dataXd), dYr.lengths)

    return Ragged(Y, Xr.lengths), backprop


def _padded_forward(
    model: Model[SeqT, SeqT], Xp: Padded, is_train: bool
) -> Tuple[Padded, Callable]:
    layer: Model[Array3d, Array3d] = model.layers[0]
    Y, get_dX = layer(Xp.data, is_train)

    def backprop(dYp: Padded) -> Padded:
        assert isinstance(dYp, Padded)
        dX = get_dX(dYp.data)
        return Padded(dX, dYp.size_at_t, dYp.lengths, dYp.indices)

    return Padded(Y, Xp.size_at_t, Xp.lengths, Xp.indices), backprop



import math
import numpy
import pytest
from thinc.api import SGD, to_categorical, SparseLinear, SparseLinear_v2


@pytest.fixture
def instances():
    lengths = numpy.asarray([5, 4], dtype="int32")
    keys = numpy.arange(9, dtype="uint64")
    values = numpy.ones(9, dtype="float32")
    X = (keys, values, lengths)
    y = numpy.asarray([0, 2], dtype="int32")
    return X, to_categorical(y, n_classes=3)


@pytest.fixture
def sgd():
    return SGD(0.001)


def test_basic(instances, sgd):
    X, y = instances
    nr_class = 3
    model = SparseLinear(nr_class).initialize()
    yh, backprop = model.begin_update(X)
    loss1 = ((yh - y) ** 2).sum()
    backprop(yh - y)
    model.finish_update(sgd)
    yh, backprop = model.begin_update(X)
    loss2 = ((yh - y) ** 2).sum()
    assert loss2 < loss1


def test_init():
    model = SparseLinear(3).initialize()
    keys = numpy.ones((5,), dtype="uint64")
    values = numpy.ones((5,), dtype="f")
    lengths = numpy.zeros((2,), dtype="int32")
    lengths[0] = 3
    lengths[1] = 2
    scores, backprop = model.begin_update((keys, values, lengths))
    assert scores.shape == (2, 3)
    d_feats = backprop(scores)
    assert len(d_feats) == 3


def test_distribution():
    n_class = 10
    length = 2**18
    model = SparseLinear_v2(nO=n_class, length=length).initialize()

    ii64 = numpy.iinfo(numpy.uint64)
    lengths = numpy.zeros((2,), dtype="int32")

    for p_nonzero in range(1, 12):
        # Clear gradients from the previous iterarion.
        model.set_grad("W", 0.0)

        n = 2**p_nonzero
        keys = numpy.random.randint(ii64.min, ii64.max, size=(n,), dtype=numpy.uint64)
        values = numpy.ones((n,), dtype="f")
        lengths[0] = n // 2
        lengths[1] = n // 2

        # Probability that a bit is set (2 because we use 2 hashes).
        p_nonzero = 1 - math.exp(-2 * n / length)

        Y, backprop = model.begin_update((keys, values, lengths))
        backprop(numpy.ones_like(Y))

        # Check that for each class we have the expected rate of non-zeros.
        dW = model.get_grad("W").reshape(n_class, -1)
        nonzero_empirical = numpy.count_nonzero(dW, axis=1) / dW.shape[1]
        numpy.testing.assert_allclose(
            nonzero_empirical, p_nonzero, rtol=1e-4, atol=1e-4
        )



from typing import Tuple, Callable, Optional, TypeVar

from ..model import Model
from ..types import ArrayXd
from ..config import registry
from ..util import get_width


LayerT = TypeVar("LayerT")
SimT = TypeVar("SimT")
InT = Tuple[LayerT, LayerT]
OutT = TypeVar("OutT", bound=ArrayXd)


@registry.layers("siamese.v1")
def siamese(
    layer: Model[LayerT, SimT], similarity: Model[Tuple[SimT, SimT], OutT]
) -> Model[InT, OutT]:
    return Model(
        f"siamese({layer.name}, {similarity.name})",
        forward,
        init=init,
        layers=[layer, similarity],
        dims={"nI": layer.get_dim("nI"), "nO": similarity.get_dim("nO")},
    )


def forward(
    model: Model[InT, OutT], X1_X2: InT, is_train: bool
) -> Tuple[OutT, Callable]:
    X1, X2 = X1_X2
    vec1, bp_vec1 = model.layers[0](X1, is_train)
    vec2, bp_vec2 = model.layers[0](X2, is_train)
    output, bp_output = model.layers[1]((vec1, vec2), is_train)

    def finish_update(d_output: OutT) -> InT:
        d_vec1, d_vec2 = bp_output(d_output)
        d_input1 = bp_vec1(d_vec1)
        d_input2 = bp_vec2(d_vec2)
        return (d_input1, d_input2)

    return output, finish_update


def init(
    model: Model[InT, OutT], X: Optional[InT] = None, Y: Optional[OutT] = None
) -> None:
    if X is not None:
        model.layers[0].set_dim("nI", get_width(X[1]))
        model.layers[0].initialize(X=X[0])
        X = (model.layers[0].predict(X[0]), model.layers[0].predict(X[1]))
    model.layers[1].initialize(X=X, Y=Y)
    model.set_dim("nI", model.layers[0].get_dim("nI"))
    model.set_dim("nO", model.layers[1].get_dim("nO"))
